// Copyright (c) .NET Foundation and contributors. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.

#include "SampleProfiler.h"
#include "corhlpr.h"
#include <string>
#include <assert.h>




#include <fstream>
#include <list>
constexpr char logfile[] = "/home/viewizard/Desktop/log.log";


using std::shared_ptr;

struct survRange
{
    std::uintptr_t rangeStart=0;
    std::uintptr_t rangeLength=0;
    survRange(std::uintptr_t start, std::uintptr_t len) : 
        rangeStart(start), rangeLength(len)
    {}
};
std::list<survRange> survRangeList;

SampleProfiler::SampleProfiler() :
    refCount(0),
    corProfilerInfo(nullptr),
    //sampler(),
    jitEventCount(0)
{

}

SampleProfiler::~SampleProfiler()
{
    if (this->corProfilerInfo != nullptr)
    {
        this->corProfilerInfo->Release();
        this->corProfilerInfo = nullptr;
    }
}

HRESULT STDMETHODCALLTYPE SampleProfiler::Initialize(IUnknown *pICorProfilerInfoUnk)
{
    printf("Initialize profiler!\n");

    HRESULT hr = pICorProfilerInfoUnk->QueryInterface(IID_ICorProfilerInfo10, (void**)&corProfilerInfo);
    if (hr != S_OK)
    {
        printf("Got HR %X from QI for ICorProfilerInfo4", hr);
        return E_FAIL;
    }

    corProfilerInfo->SetEventMask2(COR_PRF_ENABLE_FUNCTION_ARGS | COR_PRF_ENABLE_FUNCTION_RETVAL | COR_PRF_ENABLE_FRAME_INFO |
         COR_PRF_MONITOR_CLASS_LOADS | COR_PRF_ENABLE_OBJECT_ALLOCATED | COR_PRF_MONITOR_OBJECT_ALLOCATED | COR_PRF_MONITOR_GC | COR_PRF_MONITOR_JIT_COMPILATION, 0);

    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::Shutdown()
{
    if (this->corProfilerInfo != nullptr)
    {
        this->corProfilerInfo->Release();
        this->corProfilerInfo = nullptr;
    }

    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AppDomainCreationStarted(AppDomainID appDomainId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AppDomainCreationFinished(AppDomainID appDomainId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AppDomainShutdownStarted(AppDomainID appDomainId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AppDomainShutdownFinished(AppDomainID appDomainId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AssemblyLoadStarted(AssemblyID assemblyId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AssemblyLoadFinished(AssemblyID assemblyId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AssemblyUnloadStarted(AssemblyID assemblyId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::AssemblyUnloadFinished(AssemblyID assemblyId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ModuleLoadStarted(ModuleID moduleId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ModuleLoadFinished(ModuleID moduleId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ModuleUnloadStarted(ModuleID moduleId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ModuleUnloadFinished(ModuleID moduleId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ModuleAttachedToAssembly(ModuleID moduleId, AssemblyID AssemblyId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ClassLoadStarted(ClassID classId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ClassLoadFinished(ClassID classId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ClassUnloadStarted(ClassID classId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ClassUnloadFinished(ClassID classId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::FunctionUnloadStarted(FunctionID functionId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::JITCompilationStarted(FunctionID functionId, BOOL fIsSafeToBlock)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::JITCompilationFinished(FunctionID functionId, HRESULT hrStatus, BOOL fIsSafeToBlock)
{
    ++jitEventCount;
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::JITCachedFunctionSearchStarted(FunctionID functionId, BOOL *pbUseCachedFunction)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::JITCachedFunctionSearchFinished(FunctionID functionId, COR_PRF_JIT_CACHE result)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::JITFunctionPitched(FunctionID functionId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::JITInlining(FunctionID callerId, FunctionID calleeId, BOOL *pfShouldInline)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ThreadCreated(ThreadID threadId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ThreadDestroyed(ThreadID threadId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ThreadAssignedToOSThread(ThreadID managedThreadId, DWORD osThreadId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingClientInvocationStarted()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingClientSendingMessage(GUID *pCookie, BOOL fIsAsync)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingClientReceivingReply(GUID *pCookie, BOOL fIsAsync)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingClientInvocationFinished()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingServerReceivingMessage(GUID *pCookie, BOOL fIsAsync)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingServerInvocationStarted()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingServerInvocationReturned()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RemotingServerSendingReply(GUID *pCookie, BOOL fIsAsync)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::UnmanagedToManagedTransition(FunctionID functionId, COR_PRF_TRANSITION_REASON reason)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ManagedToUnmanagedTransition(FunctionID functionId, COR_PRF_TRANSITION_REASON reason)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RuntimeSuspendStarted(COR_PRF_SUSPEND_REASON suspendReason)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RuntimeSuspendFinished()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RuntimeSuspendAborted()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RuntimeResumeStarted()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RuntimeResumeFinished()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RuntimeThreadSuspended(ThreadID threadId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RuntimeThreadResumed(ThreadID threadId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::MovedReferences(ULONG cMovedObjectIDRanges, ObjectID oldObjectIDRangeStart[], ObjectID newObjectIDRangeStart[], ULONG cObjectIDRangeLength[])
{
    // should never call this on our example
    std::abort();

    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ObjectAllocated(ObjectID objectId, ClassID classId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ObjectsAllocatedByClass(ULONG cClassCount, ClassID classIds[], ULONG cObjects[])
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ObjectReferences(ObjectID objectId, ClassID classId, ULONG cObjectRefs, ObjectID objectRefIds[])
{
  HRESULT hr;

  for (int i = 0; i < cObjectRefs; ++i) {
    ClassID subjClass;
    hr = this->corProfilerInfo->GetClassFromObject(objectRefIds[i], &subjClass);

    if (hr != S_OK) {
      fprintf(stderr, "[W] Unknown type for object %" PRIxPTR ", HRESULT=%x\n", objectRefIds[i], hr);
      continue;
    }

    {
        std::ofstream outfile;
        outfile.open(logfile, std::ios_base::app);
        outfile << "ObjectReferences objectId=" << objectId << " classId=" << classId << " objectRefId=" << objectRefIds[i] << " objectRefClassId=" << subjClass << "\n";
    }

    bool found = false;
    for (auto &entry : survRangeList)
    {
        if ((objectRefIds[i] >= entry.rangeStart) && (objectRefIds[i] < entry.rangeStart + entry.rangeLength))
        {
            found = true;
            break;
        }
    }
    if (!found)
    {
        std::ofstream outfile;
        outfile.open(logfile, std::ios_base::app);
        outfile << "!!! ObjectReferences objectRefId=" << objectRefIds[i] << " not in surviving range\n";
    }

  }
  return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RootReferences(ULONG cRootRefs, ObjectID rootRefIds[])
{
  HRESULT hr;

  for (int i = 0; i < cRootRefs; ++i) {
    if (rootRefIds[i] == 0) {
      continue;
    }

    ClassID rootClass;
    hr = this->corProfilerInfo->GetClassFromObject(rootRefIds[i], &rootClass);

    if (hr != S_OK)
    {
      fprintf(stderr, "[W] Unknown type for object %" PRIxPTR ", HRESULT=%x\n", rootRefIds[i], hr);
      continue;
    }

    {
        std::ofstream outfile;
        outfile.open(logfile, std::ios_base::app);
        outfile << "RootReferences rootRefId=" << rootRefIds[i] << " rootRefIdsClassId=" << rootClass << "\n";
    }


    bool found = false;
    for (auto &entry : survRangeList)
    {
        if ((rootRefIds[i] >= entry.rangeStart) && (rootRefIds[i] < entry.rangeStart + entry.rangeLength))
        {
            found = true;
            break;
        }
    }
    if (!found)
    {
        std::ofstream outfile;
        outfile.open(logfile, std::ios_base::app);
        outfile << "!!! RootReferences objectId=" << rootRefIds[i] << " not in surviving range\n";
    }

  }
  return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionThrown(ObjectID thrownObjectId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionSearchFunctionEnter(FunctionID functionId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionSearchFunctionLeave()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionSearchFilterEnter(FunctionID functionId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionSearchFilterLeave()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionSearchCatcherFound(FunctionID functionId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionOSHandlerEnter(UINT_PTR __unused)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionOSHandlerLeave(UINT_PTR __unused)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionUnwindFunctionEnter(FunctionID functionId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionUnwindFunctionLeave()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionUnwindFinallyEnter(FunctionID functionId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionUnwindFinallyLeave()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionCatcherEnter(FunctionID functionId, ObjectID objectId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionCatcherLeave()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::COMClassicVTableCreated(ClassID wrappedClassId, REFGUID implementedIID, void *pVTable, ULONG cSlots)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::COMClassicVTableDestroyed(ClassID wrappedClassId, REFGUID implementedIID, void *pVTable)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionCLRCatcherFound()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ExceptionCLRCatcherExecute()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ThreadNameChanged(ThreadID threadId, ULONG cchName, WCHAR name[])
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::GarbageCollectionStarted(int cGenerations, BOOL generationCollected[], COR_PRF_GC_REASON reason)
{
    survRangeList.clear();

    HRESULT hr;
    ULONG numRanges;

    HRESULT hr2 = this->corProfilerInfo->GetGenerationBounds(0, &numRanges, NULL);
    if (hr2 != S_OK) {
      assert(false && "Failed to retreive number of ranges for a generation");
    }

    COR_PRF_GC_GENERATION_RANGE *ranges = new COR_PRF_GC_GENERATION_RANGE[numRanges];

    ULONG numRanges2;
    hr2 = this->corProfilerInfo->GetGenerationBounds(numRanges, &numRanges2, ranges);

    if (hr2 != S_OK) {
      assert(false && "Failed to retreive ranges for a generation");
    }

    assert(numRanges == numRanges2);

    {
        std::ofstream outfile;
        outfile.open(logfile, std::ios_base::app);
        outfile << "GarbageCollectionStarted numRanges=" << numRanges << " numRanges2=" << numRanges2 << " cGenerations=" << cGenerations << "\n";
    }


    for (ULONG i = 0; i < numRanges; i++)
    {
        {
            std::ofstream outfile;
            outfile.open(logfile, std::ios_base::app);
            outfile << "GarbageCollectionStarted ranges[i].generation=" << ranges[i].generation << " generationCollected[ranges[i].generation]=" << generationCollected[ranges[i].generation] << "\n";
            outfile << "GarbageCollectionStarted rangeStart=" << reinterpret_cast<uintptr_t>(ranges[i].rangeStart) << " rangeLength=" << ranges[i].rangeLength << "\n";
        }

        if (generationCollected[ranges[i].generation])
            continue;

        survRangeList.push_back(survRange(ranges[i].rangeStart, ranges[i].rangeLength));
    }

    delete [] ranges;
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::SurvivingReferences(ULONG cSurvivingObjectIDRanges, ObjectID objectIDRangeStart[], ULONG cObjectIDRangeLength[])
{

    for(ULONG i = 0; i < cSurvivingObjectIDRanges; i++) {
    {
        std::ofstream outfile;
        outfile.open(logfile, std::ios_base::app);
        outfile << "SurvivingReferences rangeStart=" << objectIDRangeStart[i] << " rangeLength=" << cObjectIDRangeLength[i] << "\n";
    }

        survRangeList.push_back(survRange(objectIDRangeStart[i], cObjectIDRangeLength[i]));
    }

    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::GarbageCollectionFinished()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::FinalizeableObjectQueued(DWORD finalizerFlags, ObjectID objectID)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::RootReferences2(ULONG cRootRefs, ObjectID rootRefIds[], COR_PRF_GC_ROOT_KIND rootKinds[], COR_PRF_GC_ROOT_FLAGS rootFlags[], UINT_PTR rootIds[])
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::HandleCreated(GCHandleID handleId, ObjectID initialObjectId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::HandleDestroyed(GCHandleID handleId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::InitializeForAttach(IUnknown *pcorProfilerInfoUnk, void *pvClientData, UINT cbClientData)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ProfilerAttachComplete()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ProfilerDetachSucceeded()
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ReJITCompilationStarted(FunctionID functionId, ReJITID rejitId, BOOL fIsSafeToBlock)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::GetReJITParameters(ModuleID moduleId, mdMethodDef methodId, ICorProfilerFunctionControl *pFunctionControl)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ReJITCompilationFinished(FunctionID functionId, ReJITID rejitId, HRESULT hrStatus, BOOL fIsSafeToBlock)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ReJITError(ModuleID moduleId, mdMethodDef methodId, FunctionID functionId, HRESULT hrStatus)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::MovedReferences2(ULONG cMovedObjectIDRanges, ObjectID oldObjectIDRangeStart[], ObjectID newObjectIDRangeStart[], SIZE_T cObjectIDRangeLength[])
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::SurvivingReferences2(ULONG cSurvivingObjectIDRanges, ObjectID objectIDRangeStart[], SIZE_T cObjectIDRangeLength[])
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ConditionalWeakTableElementReferences(ULONG cRootRefs, ObjectID keyRefIds[], ObjectID valueRefIds[], GCHandleID rootIds[])
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::GetAssemblyReferences(const WCHAR *wszAssemblyPath, ICorProfilerAssemblyReferenceProvider *pAsmRefProvider)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::ModuleInMemorySymbolsUpdated(ModuleID moduleId)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::DynamicMethodJITCompilationStarted(FunctionID functionId, BOOL fIsSafeToBlock, LPCBYTE ilHeader, ULONG cbILHeader)
{
    return S_OK;
}

HRESULT STDMETHODCALLTYPE SampleProfiler::DynamicMethodJITCompilationFinished(FunctionID functionId, HRESULT hrStatus, BOOL fIsSafeToBlock)
{
    return S_OK;
}

bool SampleProfiler::IsRuntimeExecutingManagedCode()
{
    return jitEventCount.load() > 0;
}
